/***************************************************************
 * Name:      audio.h
 * Purpose:   audio streams raw waveforms to audio
 * Author:    John Conner Atkeson
 * Created:   2007-11-06
 * Copyright: John Atkeson (http://jca3.freeshell.org)
 * License:   LPGL
 *  code heavily derived from:
 *  https://lampka.siedziba.pl/WebSVN/filedetails.php?repname=siedziba+svn&path=%2FSound_Test%2FSoundTest.cpp&rev=256&sc=1
 **************************************************************/
#include "SDL/SDL.h"//-lSDLmain -lSDL -mwindow
#include "SDL/SDL_audio.h"

#ifndef audio_h
#define audio_h 1
#include "ersatz.h"

#include <iostream>
using namespace std;
/* ************************************************************************************************************ */
/* Playable_Class is the base class for every object that can output a waveform. */
class Playable_Class{
public:
  //double xloc, yloc, zloc;// time, octave, and volume
  double time, loudness;
  double frequency, octave;

  Playable_Class::Playable_Class();
  void Playable_Class::set_octave(double new_octave);
  void Playable_Class::set_frequency(double new_frequency);
  virtual double Playable_Class::getamp(double time);
  virtual double Playable_Class::maxamp();
  virtual double Playable_Class::get_duration();// tells how long this track will play
  virtual bool Playable_Class::keepgoing();// tells when we have passed the end of duration.
  virtual class Time_Slicer_Class* Playable_Class::spawn_slicer();// creates a "read head" to keep a time marker in this playable
};/* Playable_Class */
/* ************************************************************************************************************ */
/* Time_Slicer_Class is like the "read head" for playing each playable class.
 Each descendent of Playable_Class defines & creates its own descendent of Time_Slicer_Class.
 It forks into a tree when the playable branches too, which is what the subslicers are for. */
class Time_Slicer_Class{
public:
  class Playable_Class *my_playable;
  bool keeprunning;
  ArrayList<Time_Slicer_Class> *subslicers;
  //-------------------------------------------------------------------------------------
  Time_Slicer_Class::Time_Slicer_Class(class Playable_Class *new_playable){
    my_playable = new_playable;
    subslicers = new ArrayList<Time_Slicer_Class>();
    keeprunning = true;
  };
  virtual Time_Slicer_Class::~Time_Slicer_Class(){
    delete subslicers;
  };
  virtual void Time_Slicer_Class::init_slicer(int new_start_dex, int new_end_dex){// this one is different for every descendent
  };
  virtual double Time_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
    double returnval = 0.0;
    nowtime -= my_playable->time; nowoctave += my_playable->octave;/* yloc is base octave for this frame */ nowvol *= my_playable->loudness;
    int num_subslices = subslicers->size();
    for (int scnt=0;scnt<num_subslices;scnt++){
      Time_Slicer_Class *slice = subslicers->get(scnt);
      returnval += slice->getslice(nowtime, nowoctave, nowvol);
    }
    // if we have passed the end of my playable's duration, set keepgoing to false.
    // keepgoing = my_playable->keepgoing();
    return returnval;
  };
  virtual bool Time_Slicer_Class::keepgoing(){// tells when we have passed the end of duration.
    return keeprunning;
  };
};/* Time_Slicer_Class */
/* ************************************************************************************************************ */
struct stereo{
  short left, right; /* left channel, right channel */
};
/* ************************************************************************************************************ */
class Player_Class{
  // User controllable variables
  unsigned  SAMPLE_FREQ;     // sampling frequency
  unsigned  sdl_samples;          // SDL buffer size

  float sample_rate;
  int samples_per_chunk;
  int last_sample_index;
  int bufsize;

public:
  unsigned Initialised;
  // float Frequency;// frequency of the output sound - start at middle C
  double timeinc;
  double time;
  // class Signal_Generator_Class *sgen;
  class Time_Slicer_Class *slicer;// the root slicer
  //-------------------------------------------------------------------------------------
  Player_Class::Player_Class(){
    Initialised = 0;
    // Frequency = 262.0f;       // frequency of the output sound - start at middle C
    this->SAMPLE_FREQ = 44100;     // sampling frequency
    this->sdl_samples = 2048;          // SDL buffer size
    sample_rate = 44100;// samples per second
    samples_per_chunk = ((int)(sample_rate/30.0));// this works on
    last_sample_index = samples_per_chunk-1;
    bufsize = 4 * samples_per_chunk;
    timeinc = 1.0/sample_rate;
    // CyclePos = 0.0;                     // position through the cycle
    time=0.0;
  }
  //-------------------------------------------------------------------------------------
  virtual Player_Class::~Player_Class(){
    if (Initialised){
      quit();
    }
    delete slicer;
  }
  //-------------------------------------------------------------------------------------
  void Player_Class::sdlsoundinit();// initializes sound output
  //-------------------------------------------------------------------------------------
  void Player_Class::quit(){
    SDL_QuitSubSystem(SDL_INIT_AUDIO);/* Uninitialization */
    Initialised = false;
  }
  //-------------------------------------------------------------------------------------
  void Player_Class::add_moment(double left_channel, double right_channel){
    // this would require a second buffer
  }
  //-------------------------------------------------------------------------------------
  virtual int Player_Class::fill_outbuffer(stereo *fill, int playsize);// the threaded callback
  //-------------------------------------------------------------------------------------
  int Player_Class::fill_outbuffer_sgen(stereo *fill, int playsize);
  // int Player_Class::single_slice(stereo *fill, int playsize){
  // maybe these should be in descendent classes?
//  these need to launch separate threads, or at least;
  //void Player_Class::play_signal_generator_loop(class Signal_Generator_Class *sgen);
  //void Player_Class::play_timeline_loop(class Voice_Track_Class *Voice_Track, int startdex, int enddex);
};/* Player_Class */
/* ************************************************************************************************************ */
class Signal_Generator_Class{// : public Playable_Class{
public:
  double freq;
  double octave;
  double loudness;// range 0.0 to 1.0

  double prev_time;
  double current_angle;// double CyclePos;// position through the cycle
  //-------------------------------------------------------------------------------------
  Signal_Generator_Class::Signal_Generator_Class();
  virtual Signal_Generator_Class::~Signal_Generator_Class(){
  }
  //-------------------------------------------------------------------------------------
  virtual double Signal_Generator_Class::getamp(double time);
  //-------------------------------------------------------------------------------------
  double Signal_Generator_Class::maxamp();
  //-------------------------------------------------------------------------------------
  virtual void Signal_Generator_Class::clear();
  //-------------------------------------------------------------------------------------
  virtual void Signal_Generator_Class::reset_time(double new_time);
  //-------------------------------------------------------------------------------------
  void Signal_Generator_Class::set_octave(double new_octave);
  //-------------------------------------------------------------------------------------
  void Signal_Generator_Class::set_frequency(double new_freq);
  //-------------------------------------------------------------------------------------
protected:
  //-------------------------------------------------------------------------------------
  void calc_angle(double time);
};/* Signal_Generator_Class */
/* ************************************************************************************************************ */
class Freewave_Generator_Class : public Signal_Generator_Class{// freeform wave pattern
public:
  int rez; double frez;// resolution of our waveform
  double *ampray; // sample array of 1 cycle
  Freewave_Generator_Class::Freewave_Generator_Class(double new_freq);
  virtual Freewave_Generator_Class::~Freewave_Generator_Class();
  overrider double Freewave_Generator_Class::getamp(double time);// time is in absolute seconds from T=0.
};/* Freewave_Generator_Class */
/* ************************************************************************************************************ */
class Splinewave_Generator_Class : public Freewave_Generator_Class{// freeform wave pattern
public:
  class Point_Class {
  public:
    double xloc,yloc;
    Point_Class::Point_Class(double newx,double newy){
      xloc=newx; yloc=newy;
    }
  };
  ArrayList<Point_Class> *spline;
  // double xp0, xp1, xp2; double yp0, yp1, yp2;
  //-------------------------------------------------------------------------------------
  Splinewave_Generator_Class::Splinewave_Generator_Class(double new_freq);
  //-------------------------------------------------------------------------------------
  virtual Splinewave_Generator_Class::~Splinewave_Generator_Class(){
    int num_points = spline->size();
    for (int cnt=0;cnt<num_points;cnt++){
      delete spline->get(cnt);
    }
    spline->clear();
    delete spline;
  };
  //-------------------------------------------------------------------------------------
  void Splinewave_Generator_Class::apply_spline();// apply the spline to the amplitude array
  //-------------------------------------------------------------------------------------
  // double Splinewave_Generator_Class::getamp(double time);// time is in absolute seconds from T=0.
  //-------------------------------------------------------------------------------------
  double Splinewave_Generator_Class::spline_amp(double X, double xp0, double yp0, double xp1, double yp1, double xp2, double yp2);
};/* Splinewave_Generator_Class */
/* ************************************************************************************************************ */
class Event_Marker_Class : public Playable_Class{
public:
  bool audible;
  //-------------------------------------------------------------------------------------
  class Voice_Track_Class *my_voice;
  //-------------------------------------------------------------------------------------
  Event_Marker_Class::Event_Marker_Class();
  //-------------------------------------------------------------------------------------
  void Event_Marker_Class::intep_to_next_frequency(Event_Marker_Class *nextmarker, double absolutetime, Signal_Generator_Class *sgen);
  //-------------------------------------------------------------------------------------
  void Event_Marker_Class::intep_to_next_octave(Event_Marker_Class *nextmarker, double absolutetime, Signal_Generator_Class *sgen);
  //-------------------------------------------------------------------------------------
  overrider double Event_Marker_Class::getamp(double time);
  overrider double Event_Marker_Class::maxamp();
  overrider double Event_Marker_Class::get_duration();// tells how long this track will play
  overrider bool Event_Marker_Class::keepgoing();// tells when we have passed the end of duration.
  overrider class Time_Slicer_Class* Event_Marker_Class::spawn_slicer();// creates a "read head" to keep a time marker in this playable
  //-------------------------------------------------------------------------------------
class Event_Marker_Slicer_Class : public Time_Slicer_Class{/* inner class */
    Event_Marker_Class *My_Event_Marker;
  public:
    Event_Marker_Slicer_Class::Event_Marker_Slicer_Class(class Event_Marker_Class *New_Event_Marker);
    overrider void Event_Marker_Slicer_Class::init_slicer(int new_start_dex, int new_end_dex);// this one is different for every descendent
    overrider double Event_Marker_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol);// Called once per timeslice. Must always go forward.
    overrider bool Event_Marker_Slicer_Class::keepgoing();// tells when we have passed the end of duration.
  };/* Event_Marker_Slicer_Class */
};/* Event_Marker_Class */
/* ************************************************************************************************************ */
class Voice_Track_Class : public ArrayList<Event_Marker_Class>, public Playable_Class{
public:
  double length;// duration of this track, in time
  Player_Class *my_audio;
  // Signal_Generator_Class *sgen;// for now, just one signal generator
  Signal_Generator_Class *sgen;// for now, just one signal generator
  //-------------------------------------------------------------------------------------
  Voice_Track_Class::Voice_Track_Class(Signal_Generator_Class *new_sgen, Player_Class *new_audio);// constructor
  //-------------------------------------------------------------------------------------
  Voice_Track_Class::~Voice_Track_Class();// destructor
  //-------------------------------------------------------------------------------------
  overrider void Voice_Track_Class::add(Event_Marker_Class *new_member);
  //-------------------------------------------------------------------------------------
  int find_by_time(double target_time);// given the time, treesearch and return the closest index
  //-------------------------------------------------------------------------------------
  void fill_test();// fill Voice_Track with dummy values
  //-------------------------------------------------------------------------------------
  void play_segment(int start_dex, int end_dex);
  /* cloning?
   http://www.new-brunswick.net/workshop/c++/faq/virtual-functions.html
   Voice_Track_Class* Voice_Track_Class::clone()  const { return new Voice_Track_Class(*this); }
  */
  //-------------------------------------------------------------------------------------
  /*

  oy, now we have to decide if we want to remove getamp from the playable class, and give it exclusively to the player head (timeslicer) class.
  that would mean giving signal generator its own player head class too wouldn't it?  unless sgen IS a playerhead class.

  but playerheads contain all of the continuity info for playing.

  perhaps timeline_class getamp() can call its player/slicer from inside?

  */
  /* Playable_Class members */
#if 0
  double Voice_Track_Class::getamp(double time){// Called once per timeslice. Must always go forward.
    double returnval = 0.0;
    if (time >= event1->time){// iterators
      event0 = event1;
      event1 = this->get(current_event_dex);
      current_event_dex++;
      if (current_event_dex>end_dex){
        // then one full iteration has passed. now jump out.
        keeprunning = false;
        /*
        double timespan = this->get_duration();
        bool keepon = this->keepgoing();
        */
      }
    }else{// meat of computation
      event0->intep_to_next_octave(event1, time, sgen);// morph the signal generator's frequency and loudness
      returnval = sgen->getamp(time);
      // returnval = event0->intep_to_next_octave(event1, time, sgen);// morph the signal generator's frequency and loudness
    }
    return returnval;
  };
#endif
  //-------------------------------------------------------------------------------------
  double Voice_Track_Class::maxamp(){
    return 1.0;
  }
  //-------------------------------------------------------------------------------------
  overrider double Voice_Track_Class::get_duration();// tells how long this track will play
  overrider bool Voice_Track_Class::keepgoing();// tells when we have passed the end of duration.
  overrider Time_Slicer_Class* Voice_Track_Class::spawn_slicer();// creates a "read head" to keep a time marker in this playable
  /* ************************************************************************************************************ */
class Voice_Track_Slicer_Class : public Time_Slicer_Class{/* inner class */
  public:
    Voice_Track_Class *Voice_Track;
    int current_event_dex;// do we store continuity values in tmeline?  in the short term I guess we do.
    Event_Marker_Class *event0, *event1;
    Signal_Generator_Class *sgen;// signal generator
    int start_dex, end_dex;
    bool keeprunning;
    double local_time;
    //-------------------------------------------------------------------------------------
    Voice_Track_Slicer_Class::Voice_Track_Slicer_Class(Voice_Track_Class *new_timeline);
    //-------------------------------------------------------------------------------------
    void Voice_Track_Slicer_Class::init_slicer(int new_start_dex, int new_end_dex);
    //-------------------------------------------------------------------------------------
    double Voice_Track_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol);// Called once per timeslice. Must always go forward.
    //-------------------------------------------------------------------------------------
    /* Playable_Class members */
    double Voice_Track_Slicer_Class::getamp(double time);// Called once per timeslice. Must always go forward.
    /*
    anyway no matter what if you want to hide the stereo *fill array, you'll need another array yes?
    *fill only exists inside of fill_outbuffer so nobody else can write to it.

    hm, unless
    global *fillpter
    global int cnt;

    callback(stream *fill){
    fillptr = fill;
    cnt=0;
    method fillbuffer();
    }
    add_moment(left, right){
    fill[cnt]=left, right
    cnt++;
    }
    fillbuffer(){
    add_moment(left, right);
    }
    //------------------------------------------------------
    event *event0, *event1;
    void init_timeslicer(int startdex, int enddex){
    event0 = Voice_Track[startdex];
    event1 = Voice_Track[enddex];
    cnt = 1;
    }
    bool timeslicer(double time){// called once per timeslice
    if (subfunction(time)){// subfunction returns true when done.
      double amp = event0.interpolate(event1, time);
      if (at end of event1){
        event0 = event1;
        event1 = Voice_Track[cnt];
        cnt++;
      }
    }
    return true;
    }Signal_Generator_Class
    //------------------------------------------------------
    */
  };/* Voice_Track_Slicer_Class */
};/* Voice_Track_Class */
/* ************************************************************************************************************ */
/* Group_Class can transpose its contents in pitch and volume. */
class Group_Class:public Playable_Class{
public:
  double duration, octave, amplitude;// transposing/translation values
  Playable_Class *contents;
  //-------------------------------------------------------------------------------------
  Group_Class::Group_Class(){
    duration=1.0; octave=5.0; amplitude=1.0;
    contents = NULL;
  }
  virtual Group_Class::~Group_Class(){
  }
  //-------------------------------------------------------------------------------------
  /* Playable_Class members */
  double Group_Class::getamp(double time){
    double subamp = contents->getamp(time);
    return 0.0;
  }
  //-------------------------------------------------------------------------------------
  double Group_Class::maxamp(){
    return amplitude;
  }
  //-------------------------------------------------------------------------------------
  overrider double Group_Class::get_duration(){// tells how long this track will play
    return duration;
  }
  overrider bool Group_Class::keepgoing(){// tells when we have passed the end of duration.
    return true;
  }
  overrider Time_Slicer_Class* Group_Class::spawn_slicer(){// creates a "read head" to keep a time marker in this playable
    return NULL;
  }
};/* Group_Class */
/* ************************************************************************************************************ */
/* Every branch of Time_Vine_Class launches a parallel narrative (transposed by the element's internal rules?) */
class Time_Vine_Class:public Playable_Class{
public:
  ArrayList<Playable_Class> *branches;
  Time_Vine_Class::Time_Vine_Class(){
    branches = new ArrayList<Playable_Class>();
  }
  virtual Time_Vine_Class::~Time_Vine_Class(){
    delete branches;
  }
  //-------------------------------------------------------------------------------------
  /* Playable_Class members */
  double getamp(double time){
    return 0.0;
  }
  double maxamp(){
    return 1.0;
  }
  overrider double Time_Vine_Class::get_duration(){// tells how long this track will play
    return 1.0;
  }
  overrider bool Time_Vine_Class::keepgoing(){// tells when we have passed the end of duration.
    return true;
  }
  overrider Time_Slicer_Class* Time_Vine_Class::spawn_slicer(){// creates a "read head" to keep a time marker in this playable
    return NULL;
  }
};/* Time_Vine_Class */
/* ************************************************************************************************************ */

// namespace myNamespace{int a, b;}
interface iface{// is this anything like interface in java?
public:
  void runit();
};
class testclass:public iface{
};

#endif // audio_h

#if 0 // junkyard
/*
slicer;
pro            con;
1 buffer       fancy conditionals to run right piece of code.
less latency;
stops on a dime;

loop:
pro:           con;
structured;    2 buffers
java-portable  extra thread;
can also stop;
*/

// (p0 - p1 + ((p1 * p1) - (2.0 * X * p1) - (p0 * p2) + (p0 * X) + (p2 * X))^0.5 )/(xp0 + xp2 - (2.0 * xp1))
// simplified: (p0 - p1 + (p1^2 - p0*p2 - 2*p1*X + (p0 + p2)*X)^0.5)  /  (p0 - 2*p1 + p2)
/*

(p0 - p1 + ((p1 * p1) - (2 * X * p1) - (p0 * p2) + (p0 * X) + (p2 * X))^0.5 )/(p0 + p2 - (2 * p1))

  comp_t = 1.0 - t;
  Y =
(
((p0 - p1 + ((p1 * p1) - (2 * X * p1) - (p0 * p2) + (p0 * X) + (p2 * X))^0.5 )/(p0 + p2 - (2 * p1)))^2
 * y0)
+ (2.0
* ((p0 - p1 + ((p1 * p1) - (2 * X * p1) - (p0 * p2) + (p0 * X) + (p2 * X))^0.5 )/(p0 + p2 - (2 * p1)))
* (1-(p0 - p1 + ((p1 * p1) - (2 * X * p1) - (p0 * p2) + (p0 * X) + (p2 * X))^0.5 )/(p0 + p2 - (2 * p1)))
 * y1)
+ (
((p0 - p1 + ((p1 * p1) - (2 * X * p1) - (p0 * p2) + (p0 * X) + (p2 * X))^0.5 )/(p0 + p2 - (2 * p1)))^2
 * y2);

Y = ((p0 - 1.*p1 + (p1^2 - p0*p2 - 2*p1*X + (p0 + p2)*X)^0.5)*(2.*p2*y1 + p1*(-1.*y0 - 2.*y1 - 1.*y2) + p0*(1.*y0 + 1.*y2) + (p1^2 - p0*p2 - 2*p1*X + (p0 + p2)*X)^0.5*(1.*y0 - 2.*y1 + 1.*y2)))/(p0 - 2.*p1 + p2)^2

else{
    denom = Math.Sqrt(xp0 - 2.0*xp1 + xp2);
    if (denom==0.0) {
      xp0+=fudge;// xp0+=0.00000000000001;
      denom = fudge;
    }
    Y = ((xp0 - 1.0*xp1 + Math.Sqrt(xp1*xp1 - xp0*xp2 - 2*xp1*X + (xp0 + xp2)*X))*(2.*xp2*yp1 + xp1*(-1.0*yp0 - 2.0*yp1 - 1.0*yp2) + xp0*(1.0*yp0 + 1.0*yp2) + Math.Sqrt(xp1*xp1 - xp0*xp2 - 2.0*xp1*X + (xp0 + xp2)*X)*(1.*yp0 - 2.0*yp1 + 1.*yp2)))
        /denom;
  }

*/
#endif
